#include "SDL.h"
#include <math.h>

#include "osint.h"
#include "e8910.h"
#include "vecx.h"

#include "defines.h"

#include "common.h"

#ifdef NSPIRE
#include <os.h>
#include <libndls.h>
#endif

#define EMU_TIMER 20 /* the emulators heart beats at 20 milliseconds */

extern char* currentdir;

static SDL_Surface *screen = NULL;
static SDL_Surface *overlay_original = NULL;
static long scl_factor;
static long offx;
static long offy;

unsigned char alpha_overlay;

extern char filename_without_extensions[MAX_LENGH];

void setPixel(unsigned char x, unsigned char y, Uint32 pixel);
void line(short x1, short y1, short x2, short y2) ;

void setPixel(unsigned char x, unsigned char y, Uint32 pixel)
{
#ifdef NSPIRE
	nSDL_SetPixel(screen,x,y,pixel);
#else
	unsigned char bpp = screen->format->BytesPerPixel;
	/* Here p is the address to the pixel we want to set */
	Uint8 *p = (Uint8 *)screen->pixels + y * screen->pitch + x * bpp;

	switch(bpp) {
	case 1:
		*p = pixel;
		break;

	case 2:
		*(Uint16 *)p = pixel;
		break;

	case 3:
		if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
			p[0] = (pixel >> 16) & 0xff;
			p[1] = (pixel >> 8) & 0xff;
			p[2] = pixel & 0xff;
		} else {
			p[0] = pixel & 0xff;
			p[1] = (pixel >> 8) & 0xff;
			p[2] = (pixel >> 16) & 0xff;
		}
		break;

	case 4:
		*(Uint32 *)p = pixel;
		break;
	}
#endif
}

void line(short x1, short y1, short x2, short y2) 
{
    signed char ix;
    signed char iy;
 
    // if x1 == x2 or y1 == y2, then it does not matter what we set here
    unsigned short delta_x = (x2 > x1?(ix = 1, x2 - x1):(ix = -1, x1 - x2)) << 1;
    unsigned short delta_y = (y2 > y1?(iy = 1, y2 - y1):(iy = -1, y1 - y2)) << 1;
 
	setPixel(x1,y1,65535);
    if (delta_x >= delta_y) {
        int error = delta_y - (delta_x >> 1);        // error may go below zero
        while (x1 != x2) {
            if (error >= 0) {
                if (error || (ix > 0)) {
                    y1 += iy;
                    error -= delta_x;
                }                           // else do nothing
         }                              // else do nothing
            x1 += ix;
            error += delta_y;
            setPixel(x1,y1,65535);
        }
    } else {
        int error = delta_x - (delta_y >> 1);      // error may go below zero
        while (y1 != y2) {
            if (error >= 0) {
                if (error || (iy > 0)) {
                    x1 += ix;
                    error -= delta_y;
                }                           // else do nothing
            }                              // else do nothing
            y1 += iy;
            error += delta_x;  
            setPixel(x1,y1,65535);
        }
    }
}

void osint_render(void)
{
	SDL_FillRect(screen, NULL, 0);

	int v;
	for(v = 0; v < vector_draw_cnt; v++)
	{
		line(
				(vectors_draw[v].x0 / scl_factor), 
				(vectors_draw[v].y0 / scl_factor), 
				(vectors_draw[v].x1 / scl_factor), 
				(vectors_draw[v].y1 / scl_factor) );
	}
	
	if (overlay_original)
	{
		SDL_Rect dest_rect = {0, 0, 0, 0};
		SDL_BlitSurface(overlay_original, NULL, screen, &dest_rect);
	}

	SDL_Flip(screen);
}

static char *romfilename = "rom.dat.tns";
static char *cartfilename = NULL;

static void init()
{
	FILE *f;
	if(!(f = fopen(romfilename, "rb")))
	{
		perror(romfilename);
		exit(EXIT_FAILURE);
	}
	if(fread(rom, 1, sizeof (rom), f) != sizeof (rom))
	{
		printf("Invalid rom length\n");
		exit(EXIT_FAILURE);
	}
	fclose(f);

	memset(cart, 0, sizeof (cart));
	if(cartfilename)
	{
		FILE *f;
		if(!(f = fopen(cartfilename, "rb")))
		{
			perror(cartfilename);
			exit(EXIT_FAILURE);
		}
		fread(cart, 1, sizeof (cart), f);
		fclose(f);
	}
}

void resize(int width, int height)
{
	long sclx, scly;

	long screenx = width;
	long screeny = height;
	
#ifdef NSPIRE
	screen = SDL_SetVideoMode(screenx, screeny, has_colors ? 16 : 8, SDL_SWSURFACE);
#else
	screen = SDL_SetVideoMode(screenx, screeny, 16, SDL_SWSURFACE);
#endif

	sclx = ALG_MAX_X / screen->w;
	scly = ALG_MAX_Y / screen->h;

	scl_factor = sclx > scly ? sclx : scly;

	offx = (screenx - ALG_MAX_X / scl_factor) / 2;
	offy = (screeny - ALG_MAX_Y / scl_factor) / 2;
}

static void readevents()
{
	SDL_Event e;
	while(SDL_PollEvent(&e)){
		switch(e.type){
			case SDL_QUIT:
				exit(EXIT_SUCCESS);
				break;
				break;
			case SDL_KEYDOWN:
				switch(e.key.keysym.sym){
					case SDLK_ESCAPE:
						exit(EXIT_SUCCESS);
					case SDLK_LCTRL:
						snd_regs[14] &= ~0x01;
						break;
					case SDLK_LSHIFT:
						snd_regs[14] &= ~0x02;
						break;
					case SDLK_BACKSPACE:
						snd_regs[14] &= ~0x04;
						break;
					case SDLK_TAB:
						snd_regs[14] &= ~0x08;
						break;
					case SDLK_LEFT:
						alg_jch0 = 0x00;
						break;
					case SDLK_RIGHT:
						alg_jch0 = 0xff;
						break;
#ifdef NSPIRE
					case SDLK_MINUS:
							alpha_overlay = alpha_overlay - 16;
							if (alpha_overlay < 16) alpha_overlay = 16;
							if (overlay_original) SDL_SetAlpha(overlay_original, SDL_SRCALPHA|SDL_ACCELERATION_RLE, alpha_overlay);
						break;
					case SDLK_PLUS:
							alpha_overlay = alpha_overlay + 16;
							if (alpha_overlay > 196) alpha_overlay = 196;
							if (overlay_original) SDL_SetAlpha(overlay_original, SDL_SRCALPHA|SDL_ACCELERATION_RLE, alpha_overlay);
						break;
#else
					case SDLK_b:
							alpha_overlay = alpha_overlay - 16;
							if (alpha_overlay < 16) alpha_overlay = 16;
							if (overlay_original) SDL_SetAlpha(overlay_original, SDL_SRCALPHA|SDL_ACCELERATION_RLE, alpha_overlay);
						break;
					case SDLK_n:
							alpha_overlay = alpha_overlay + 16;
							if (alpha_overlay > 196) alpha_overlay = 196;
							if (overlay_original) SDL_SetAlpha(overlay_original, SDL_SRCALPHA|SDL_ACCELERATION_RLE, alpha_overlay);
						break;
#endif

					case SDLK_UP:
						alg_jch1 = 0xff;
						break;
					case SDLK_DOWN:
						alg_jch1 = 0x00;
						break;
					default:
						break;
				}
				break;
			case SDL_KEYUP:
				switch(e.key.keysym.sym){
					case SDLK_LCTRL:
						snd_regs[14] |= 0x01;
						break;
					case SDLK_LSHIFT:
						snd_regs[14] |= 0x02;
						break;
					case SDLK_BACKSPACE:
						snd_regs[14] |= 0x04;
						break;
					case SDLK_TAB:
						snd_regs[14] |= 0x08;
						break;
					case SDLK_LEFT:
						alg_jch0 = 0x80;
						break;
					case SDLK_RIGHT:
						alg_jch0 = 0x80;
						break;
					case SDLK_UP:
						alg_jch1 = 0x80;
						break;
					case SDLK_DOWN:
						alg_jch1 = 0x80;
						break;
#ifdef NSPIRE
					case SDLK_MINUS:
						break;
					case SDLK_PLUS:
						break;
#else
					case SDLK_b:
						break;
					case SDLK_n:
						break;
#endif
					default:
						break;
				}
				break;
			default:
				break;
		}
	}
}

void osint_emuloop()
{
#ifdef FAST
	int waiting = (VECTREX_MHZ / 1000) * EMU_TIMER;
	vecx_reset();
	
	for(;;)
	{
		vecx_emu(waiting);
		readevents();
	}
#else
	int waiting = (VECTREX_MHZ / 1000) * EMU_TIMER;
	Uint32 next_time = SDL_GetTicks() + EMU_TIMER;
	vecx_reset();
	
	for(;;)
	{
		vecx_emu(waiting);
		readevents();

		{
			Uint32 now = SDL_GetTicks();
			if(now < next_time)
				SDL_Delay(next_time - now);
			else
				next_time = now;
			next_time += EMU_TIMER;
		}
	}
#endif
}

int load_overlay(const char *filename)
{
	
	SDL_Surface *image;
	SDL_Surface *tmp;
	
	tmp = SDL_LoadBMP(filename);
	if (tmp)
	{
		image = SDL_DisplayFormat(tmp);
		SDL_FreeSurface(tmp);
	}
	else
	{
		SDL_FreeSurface(tmp);
		return 0;
	}
	
	alpha_overlay = 64;
	
	SDL_SetAlpha(image, SDL_SRCALPHA|SDL_ACCELERATION_RLE, alpha_overlay);
	overlay_original = image;
	
	return 1;
}

int vecx(char* rom_name)
{
	char temp_image_name[511];

#ifdef NOSOUND
	SDL_Init(SDL_INIT_VIDEO);
#else
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0){
		fprintf(stderr, "Failed to initialize SDL: %s\n", SDL_GetError());
		exit(-1);
	}
#endif

	resize(200, 240);
	
	cartfilename = rom_name;
	
#ifdef NSPIRE
	snprintf(temp_image_name, sizeof(temp_image_name), "%s/%s.bmp.tns", currentdir, filename_without_extensions);
#else
	snprintf(temp_image_name, sizeof(temp_image_name), "%s/%s.bmp", currentdir, filename_without_extensions);
#endif

	load_overlay(temp_image_name);

	init();
#ifdef NOSOUND
	osint_emuloop();
#else
	e8910_init_sound();
	osint_emuloop();
	e8910_done_sound();
#endif
	
	if (overlay_original) SDL_FreeSurface(overlay_original);
	SDL_Quit();

	return 0;
}

